/*
             LUFA Library
     Copyright (C) Dean Camera, 2010.
              
  dean [at] fourwalledcubicle [dot] com
      www.fourwalledcubicle.com
*/

/*
  Copyright 2010  Dean Camera (dean [at] fourwalledcubicle [dot] com)

  Permission to use, copy, modify, distribute, and sell this 
  software and its documentation for any purpose is hereby granted
  without fee, provided that the above copyright notice appear in 
  all copies and that both that the copyright notice and this
  permission notice and warranty disclaimer appear in supporting 
  documentation, and that the name of the author not be used in 
  advertising or publicity pertaining to distribution of the 
  software without specific, written prior permission.

  The author disclaim all warranties with regard to this
  software, including all implied warranties of merchantability
  and fitness.  In no event shall the author be liable for any
  special, indirect or consequential damages or any damages
  whatsoever resulting from loss of use, data or profits, whether
  in an action of contract, negligence or other tortious action,
  arising out of or in connection with the use or performance of
  this software.
*/

/** \file
 *
 *  Main source file for the Arduino-usbisp project. This file contains the main tasks of
 *  the project and is responsible for the initial application hardware configuration.
 */

#include "Arduino-usbisp.h"

/** Circular buffer to hold data from the host before it is sent to the device via the serial port. */
RingBuff_t_l USBtoUSART_Buffer;

/** Circular buffer to hold data from the serial port before it is sent to the host. */
RingBuff_t_s USARTtoUSB_Buffer;

/** LUFA CDC Class driver interface configuration and state information. This structure is
 *  passed to all CDC Class driver functions, so that multiple instances of the same class
 *  within a device can be differentiated from one another.
 */
USB_ClassInfo_CDC_Device_t VirtualSerial_CDC_Interface =
  {
    .Config = 
      {
        .ControlInterfaceNumber         = 0,

        .DataINEndpointNumber           = CDC_TX_EPNUM,
        .DataINEndpointSize             = CDC_TXRX_EPSIZE,
        .DataINEndpointDoubleBank       = false,

        .DataOUTEndpointNumber          = CDC_RX_EPNUM,
        .DataOUTEndpointSize            = CDC_TXRX_EPSIZE,
        .DataOUTEndpointDoubleBank      = false,

        .NotificationEndpointNumber     = CDC_NOTIFICATION_EPNUM,
        .NotificationEndpointSize       = CDC_NOTIFICATION_EPSIZE,
        .NotificationEndpointDoubleBank = false,
      },
  };

volatile uint8_t time_usb=0; // millisecond counter for USB
volatile uint8_t time_isp=0; // millisecond counter for ISP
/** Pulse generation counters to keep track of the number of milliseconds remaining for each pulse type */
volatile uint8_t TxLEDPulse; /**< Milliseconds remaining for data Tx LED pulse */
volatile uint8_t RxLEDPulse; /**< Milliseconds remaining for data Rx LED pulse */

#include "isp3.c"

/** Main program entry point. This routine contains the overall program flow, including initial
 *  setup of all components and the main program loop.
 */
int main(void) {
  SetupHardware();
  
  RingBuffer_InitBuffer_l(&USBtoUSART_Buffer); // large receive buffer (needed by ISP)
  RingBuffer_InitBuffer_s(&USARTtoUSB_Buffer); // smaller transmit buffer (memory constraints)

  sei();
  
  for (;;) {

    // serial mode
    AVR_RESET_LINE_DDR  |= AVR_RESET_LINE_MASK; // output
    for (;;) {
      if ( (~PINB) & (1 << JUMPER_SENSE) ) { break; } // continue whilst pin high
      avrserial();
    }
    
    // ISP mode
    AVR_RESET_LINE_DDR  &= ~AVR_RESET_LINE_MASK; // input
    for (;;) {
      if ( PINB & (1 << JUMPER_SENSE) ) { break; } // continue whilst pin low
      avrisp();
    }
  }

}

void avrserial(void) {
  service_usb();

  if (!(RingBuffer_IsEmpty_l(&USBtoUSART_Buffer))) {
    Serial_TxByte(RingBuffer_Remove_l(&USBtoUSART_Buffer));
  }
}

void service_usb(void) {
/* This routine needs to be called regularly, but is not suitable for hanging off an interrupt (?)
   That's ok when combined with short tasks in the main loop, as in serial mode, but avrisp blocks whilst reading & writing data.
   Even if avrisp returns quickly, it needs to service the USB routines in order to read & write new data
   The solution is to call service_usb() inside any routines which waits.  These are:
     delay()
     getch()
     spi_transaction()
*/
  
  // receive from USB
  if (!(RingBuffer_IsFull_l(&USBtoUSART_Buffer))) {
    int16_t ReceivedByte = CDC_Device_ReceiveByte(&VirtualSerial_CDC_Interface);
    if (!(ReceivedByte < 0)) { // not an error
      // RX led on:
      LEDs_TurnOnLEDs(LEDMASK_RX);
      cli();
      RxLEDPulse = TX_RX_LED_PULSE_MS;
      sei();
      // fill buffer
      RingBuffer_Insert_l(&USBtoUSART_Buffer, ReceivedByte);
    }
  }

  // send to USB
  RingBuff_Count_t_s BufferCount = RingBuffer_GetCount_s(&USARTtoUSB_Buffer);
  cli();
  if ((time_usb>3) || (BufferCount > BUFFER_NEARLY_FULL_S)) {
    time_usb=0;
    sei();
    if (BufferCount) { // buffer contains something
      // TX led on:
      LEDs_TurnOnLEDs(LEDMASK_TX);
      cli();
      TxLEDPulse = TX_RX_LED_PULSE_MS;
      sei();
      // empty buffer:
      while (BufferCount--) {
        CDC_Device_SendByte(&VirtualSerial_CDC_Interface, RingBuffer_Remove_s(&USARTtoUSB_Buffer));
      }
    }
  }
  sei();

  // service USB
  CDC_Device_USBTask(&VirtualSerial_CDC_Interface);
  USB_USBTask();

}

/** Configures the board hardware and chip peripherals for the demo's functionality. */
void SetupHardware(void) {
  /* Disable watchdog if enabled by bootloader/fuses */
  MCUSR &= ~(1 << WDRF);
  wdt_disable();

  /* Hardware Initialization */
  Serial_Init(9600, false);
  LEDs_Init();
  USB_Init();

  /* Start the flush timer so that overflows occur rapidly to push received bytes to the USB interface */
  TCCR0B = ((1 << CS01)|(1 << CS00)); // clk/64 (16MHz/64/256 = ~1kHz)
  TIMSK0 = (1 << TOIE0); // interrupt on overflow
  
  /* Pull target /RESET line high (pin PD7) */
  AVR_RESET_LINE_PORT |= AVR_RESET_LINE_MASK; // high
	AVR_RESET_LINE_DDR  |= AVR_RESET_LINE_MASK; // output
  
  // ISP pins
  DDRB  &=  ~( (1<<RESET) | (1<<SCK) | (1<<MOSI) | (1<<MISO) ); // inputs
  PORTB &=  ~( (1<<RESET) | (1<<SCK) | (1<<MOSI) | (1<<MISO) ); // low

  // Jumper pins
  PORTB |=  (1<<JUMPER_SENSE); // pull-up
  DDRB  &= ~(1<<JUMPER_SENSE); // input
  PORTB &= ~(1<<JUMPER_GND);   // low
  DDRB  |=  (1<<JUMPER_GND);   // output
  PORTB &= ~(1<<JUMPER_DEBUG); // low
  DDRB  |=  (1<<JUMPER_DEBUG); // output
}

/** Event handler for the library USB Configuration Changed event. */
void EVENT_USB_Device_ConfigurationChanged(void) {
  CDC_Device_ConfigureEndpoints(&VirtualSerial_CDC_Interface);
}

/** Event handler for the library USB Unhandled Control Request event. */
void EVENT_USB_Device_UnhandledControlRequest(void) {
  CDC_Device_ProcessControlRequest(&VirtualSerial_CDC_Interface);
}

/** Event handler for the CDC Class driver Line Encoding Changed event.
 *
 *  \param[in] CDCInterfaceInfo  Pointer to the CDC class interface configuration structure being referenced
 */
void EVENT_CDC_Device_LineEncodingChanged(USB_ClassInfo_CDC_Device_t* const CDCInterfaceInfo) {
  uint8_t ConfigMask = 0;

  switch (CDCInterfaceInfo->State.LineEncoding.ParityType)
  {
    case CDC_PARITY_Odd:
      ConfigMask = ((1 << UPM11) | (1 << UPM10));    
      break;
    case CDC_PARITY_Even:
      ConfigMask = (1 << UPM11);    
      break;
  }

  if (CDCInterfaceInfo->State.LineEncoding.CharFormat == CDC_LINEENCODING_TwoStopBits)
    ConfigMask |= (1 << USBS1);

  switch (CDCInterfaceInfo->State.LineEncoding.DataBits)
  {
    case 6:
      ConfigMask |= (1 << UCSZ10);
      break;
    case 7:
      ConfigMask |= (1 << UCSZ11);
      break;
    case 8:
      ConfigMask |= ((1 << UCSZ11) | (1 << UCSZ10));
      break;
  }

  /* Must turn off USART before reconfiguring it, otherwise incorrect operation may occur */
  UCSR1B = 0;
  UCSR1A = 0;
  UCSR1C = 0;

  /* Special case 57600 baud for compatibility with the ATmega328 bootloader. */  
  UBRR1  = (CDCInterfaceInfo->State.LineEncoding.BaudRateBPS == 57600)
       ? SERIAL_UBBRVAL(CDCInterfaceInfo->State.LineEncoding.BaudRateBPS)
       : SERIAL_2X_UBBRVAL(CDCInterfaceInfo->State.LineEncoding.BaudRateBPS);  

  UCSR1C = ConfigMask;
  UCSR1A = (CDCInterfaceInfo->State.LineEncoding.BaudRateBPS == 57600) ? 0 : (1 << U2X1);
  UCSR1B = ((1 << RXCIE1) | (1 << TXEN1) | (1 << RXEN1));
}

ISR(TIMER0_OVF_vect, ISR_BLOCK) {
  PORTB |= (1<<JUMPER_DEBUG); // DEBUGGING: set PB7
  // counters
  time_usb++;
  time_isp++;
  // LEDs off
  if (TxLEDPulse && !(--TxLEDPulse)) { LEDs_TurnOffLEDs(LEDMASK_TX); }
  if (RxLEDPulse && !(--RxLEDPulse)) { LEDs_TurnOffLEDs(LEDMASK_RX); }
  PORTB &= ~(1<<JUMPER_DEBUG); // DEBUGGING: clear PB7
}

/** ISR to manage the reception of data from the serial port, placing received bytes into a circular buffer
 *  for later transmission to the host.
 */
ISR(USART1_RX_vect, ISR_BLOCK) {
  uint8_t ReceivedByte = UDR1;

  if (USB_DeviceState == DEVICE_STATE_Configured) {
    RingBuffer_Insert_s(&USARTtoUSB_Buffer, ReceivedByte);
  }
}

/** Event handler for the CDC Class driver Host-to-Device Line Encoding Changed event.
 *
 *  \param[in] CDCInterfaceInfo  Pointer to the CDC class interface configuration structure being referenced
 */
void EVENT_CDC_Device_ControLineStateChanged(USB_ClassInfo_CDC_Device_t* const CDCInterfaceInfo) {
  bool CurrentDTRState = (CDCInterfaceInfo->State.ControlLineStates.HostToDevice & CDC_CONTROL_LINE_OUT_DTR);
  if (CurrentDTRState) {
    AVR_RESET_LINE_PORT &= ~AVR_RESET_LINE_MASK; // low
  } else {
    AVR_RESET_LINE_PORT |= AVR_RESET_LINE_MASK; // high
  }
}
